from cosmic_toolkit.message_bus import MessageBus
from cosmic_toolkit.models import Entity, Event
from cosmic_toolkit.repository import AbstractRepository
from cosmic_toolkit.unit_of_work import BaseUnitOfWork

__all__ = ["AbstractRepository", "BaseUnitOfWork", "Entity", "Event", "MessageBus"]



from abc import ABCMeta, abstractmethod
from typing import Generator, List

from cosmic_toolkit.models import Event
from cosmic_toolkit.repository import AbstractRepository


class BaseUnitOfWork(metaclass=ABCMeta):
    def __init__(self, *args, **kwargs):
        """Instantiate Unit of Work - arguments are passed into constructors of
        repositories"""
        # Instantiate repositories
        self._repositories = {
            k: v(*args, **kwargs) for k, v in self._repositories.items()
        }

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__()
        cls._repositories = {
            k: v for k, v in kwargs.items() if issubclass(v, AbstractRepository)
        }

    async def __aenter__(self) -> "BaseUnitOfWork":
        return self

    async def __aexit__(self, exc_type, exc, tb):
        # If transaction is committed, rollback shouldn't error
        # This is here as a fallback
        await self.rollback()

    def __getattr__(self, item: str) -> AbstractRepository:
        # Enable accessing repositories as attributes
        # E.g. uow.customers.add()
        return self._repositories[item]

    def collect_new_events(self) -> Generator[List[Event], None, None]:
        for repository in self._repositories.values():
            for entity in repository.seen:
                while entity.events:
                    yield entity.events.pop(0)

    @abstractmethod
    async def commit(self):
        ...

    @abstractmethod
    async def rollback(self):
        ...



import pytest

from cosmic_toolkit import AbstractRepository, Entity

pytestmark = pytest.mark.asyncio


def test_abstract_repository_incorrect_entity_type():
    class Item:
        ...

    with pytest.raises(TypeError) as e:

        class ItemRepository(AbstractRepository, entity_type=Item):
            async def _add(self, entity: Entity):
                ...

            async def _get(self, id: str) -> Entity:
                ...

            async def _update(self, entity: Entity):
                ...

    assert str(e.value) == "Entity must inherit from Entity"


async def test_abstract_repository_add_update_enforces_entity(
    test_entities, test_repositories
):
    entity_a = test_entities["EntityA"]()
    entity_b = test_entities["EntityB"]()
    repository_a = test_repositories["TestRepositoryA"]()

    await repository_a.add(entity_a)

    with pytest.raises(TypeError) as e:
        await repository_a.add(entity_b)

    assert str(e.value) == "Expecting entity of type EntityA"
    assert list(repository_a.seen)[0] == entity_a



from abc import ABCMeta, abstractmethod
from typing import Any, Dict, List

from pydantic import BaseModel


class Event(BaseModel):
    ...


class Entity(metaclass=ABCMeta):
    _events: List[Event]

    def __init__(self, *args, **kwargs):
        self._events = []

    @classmethod
    @abstractmethod
    def init(cls, *args, **kwargs) -> "Entity":
        ...

    @property
    def events(self) -> List[Event]:
        return self._events

    @abstractmethod
    def dict(self) -> Dict[str, Any]:
        ...



from typing import Any, Dict, Type

import pytest

from cosmic_toolkit import (
    AbstractRepository,
    BaseUnitOfWork,
    Entity,
    Event,
)


class EntityA(Entity):
    def init(cls, *args, **kwargs) -> "Entity":
        return cls()

    def dict(self) -> Dict[str, Any]:
        return {}


class EntityB(EntityA):
    def init(cls, *args, **kwargs) -> "Entity":
        return cls()

    def dict(self) -> Dict[str, Any]:
        return {}


class ATriggered(Event):
    ...


class BTriggered(Event):
    ...


class TestRepositoryA(AbstractRepository, entity_type=EntityA):
    async def _add(self, entity: EntityA):
        ...

    async def _get(self, id: str) -> EntityA:
        ...

    async def _update(self, entity: EntityA):
        ...


class TestRepositoryB(AbstractRepository, entity_type=EntityB):
    async def _add(self, entity: EntityA):
        ...

    async def _get(self, id: str) -> EntityA:
        ...

    async def _update(self, entity: EntityA):
        ...


class TestUnitOfWork(
    BaseUnitOfWork,
    a_items=TestRepositoryA,
    b_items=TestRepositoryB,
):
    async def commit(self):
        ...

    async def rollback(self):
        ...


@pytest.fixture
def test_entities() -> Dict[str, Type[Entity]]:
    return {
        "EntityA": EntityA,
        "EntityB": EntityB,
    }


@pytest.fixture
def test_events() -> Dict[str, Type[Event]]:
    return {
        "ATriggered": ATriggered,
        "BTriggered": BTriggered,
    }


@pytest.fixture
def test_repositories() -> Dict[str, Type[TestRepositoryA]]:
    return {
        "TestRepositoryA": TestRepositoryA,
        "TestRepositoryB": TestRepositoryB,
    }


@pytest.fixture
def test_unit_of_work() -> Type[TestUnitOfWork]:
    return TestUnitOfWork



import pytest

pytestmark = pytest.mark.asyncio


def test_base_unit_of_work_repository_attributes(test_unit_of_work):
    # Ensure repos can be accessed as attributes
    uow = test_unit_of_work()

    assert hasattr(uow, "a_items")
    assert hasattr(uow, "b_items")

    assert isinstance(uow.a_items.seen, set)


async def test_base_unit_of_work_collect_new_events(
    test_entities, test_events, test_unit_of_work
):
    uow = test_unit_of_work()

    # Add entities
    async with uow:
        entity_a = test_entities["EntityA"]()
        entity_a._events.append(test_events["ATriggered"]())
        entity_a._events.append(test_events["BTriggered"]())
        entity_b = test_entities["EntityB"]()
        entity_b._events.append(test_events["BTriggered"]())

        await uow.a_items.add(entity_a)
        await uow.b_items.add(entity_b)

        await uow.commit()

    # Collect events
    events = list(uow.collect_new_events())

    assert len(events) == 3
    assert isinstance(events[0], test_events["ATriggered"])
    assert isinstance(events[1], test_events["BTriggered"])
    assert isinstance(events[2], test_events["BTriggered"])
